const char schema_lua[] =
"-- schema.lua (internal file)\n"
"--\n"
"local ffi = require('ffi')\n"
"local buffer = require('buffer')\n"
"local msgpack = require('msgpack')\n"
"local msgpackffi = require('msgpackffi')\n"
"local fun = require('fun')\n"
"local log = require('log')\n"
"local fio = require('fio')\n"
"local json = require('json')\n"
"local session = box.session\n"
"local internal = require('box.internal')\n"
"local function setmap(table)\n"
"    return setmetatable(table, { __serialize = 'map' })\n"
"end\n"
"\n"
"local builtin = ffi.C\n"
"\n"
"-- performance fixup for hot functions\n"
"local tuple_encode = box.tuple.encode\n"
"local tuple_bless = box.tuple.bless\n"
"local is_tuple = box.tuple.is\n"
"assert(tuple_encode ~= nil and tuple_bless ~= nil and is_tuple ~= nil)\n"
"local cord_ibuf_take = buffer.internal.cord_ibuf_take\n"
"local cord_ibuf_put = buffer.internal.cord_ibuf_put\n"
"\n"
"local INT64_MIN = tonumber64('-9223372036854775808')\n"
"local INT64_MAX = tonumber64('9223372036854775807')\n"
"\n"
"ffi.cdef[[\n"
"    struct space *space_by_id(uint32_t id);\n"
"    extern uint32_t box_schema_version();\n"
"    void space_run_triggers(struct space *space, bool yesno);\n"
"    size_t space_bsize(struct space *space);\n"
"\n"
"    typedef struct tuple box_tuple_t;\n"
"    typedef struct iterator box_iterator_t;\n"
"\n"
"    /** \\cond public */\n"
"    box_iterator_t *\n"
"    box_index_iterator(uint32_t space_id, uint32_t index_id, int type,\n"
"                       const char *key, const char *key_end);\n"
"    int\n"
"    box_iterator_next(box_iterator_t *itr, box_tuple_t **result);\n"
"    void\n"
"    box_iterator_free(box_iterator_t *itr);\n"
"    /** \\endcond public */\n"
"    /** \\cond public */\n"
"    ssize_t\n"
"    box_index_len(uint32_t space_id, uint32_t index_id);\n"
"    ssize_t\n"
"    box_index_bsize(uint32_t space_id, uint32_t index_id);\n"
"    int\n"
"    box_index_random(uint32_t space_id, uint32_t index_id, uint32_t rnd,\n"
"                     box_tuple_t **result);\n"
"    int\n"
"    box_index_get(uint32_t space_id, uint32_t index_id, const char *key,\n"
"                  const char *key_end, box_tuple_t **result);\n"
"    int\n"
"    box_index_min(uint32_t space_id, uint32_t index_id, const char *key,\n"
"                  const char *key_end, box_tuple_t **result);\n"
"    int\n"
"    box_index_max(uint32_t space_id, uint32_t index_id, const char *key,\n"
"                  const char *key_end, box_tuple_t **result);\n"
"    ssize_t\n"
"    box_index_count(uint32_t space_id, uint32_t index_id, int type,\n"
"                    const char *key, const char *key_end);\n"
"    /** \\endcond public */\n"
"    /** \\cond public */\n"
"    bool\n"
"    box_txn();\n"
"    int64_t\n"
"    box_txn_id();\n"
"    int\n"
"    box_txn_begin();\n"
"    /** \\endcond public */\n"
"    typedef struct txn_savepoint box_txn_savepoint_t;\n"
"\n"
"    box_txn_savepoint_t *\n"
"    box_txn_savepoint();\n"
"\n"
"    struct port_tuple_entry {\n"
"        struct port_tuple_entry *next;\n"
"        struct tuple *tuple;\n"
"    };\n"
"\n"
"    struct port_tuple {\n"
"        const struct port_vtab *vtab;\n"
"        int size;\n"
"        struct port_tuple_entry *first;\n"
"        struct port_tuple_entry *last;\n"
"        struct port_tuple_entry first_entry;\n"
"    };\n"
"\n"
"    void\n"
"    port_destroy(struct port *port);\n"
"\n"
"    int\n"
"    box_select(uint32_t space_id, uint32_t index_id,\n"
"               int iterator, uint32_t offset, uint32_t limit,\n"
"               const char *key, const char *key_end,\n"
"               struct port *port);\n"
"\n"
"    void password_prepare(const char *password, int len,\n"
"                          char *out, int out_len);\n"
"\n"
"    enum priv_type {\n"
"        PRIV_R = 1,\n"
"        PRIV_W = 2,\n"
"        PRIV_X = 4,\n"
"        PRIV_S = 8,\n"
"        PRIV_U = 16,\n"
"        PRIV_C = 32,\n"
"        PRIV_D = 64,\n"
"        PRIV_A = 128,\n"
"        PRIV_REFERENCE = 256,\n"
"        PRIV_TRIGGER = 512,\n"
"        PRIV_INSERT = 1024,\n"
"        PRIV_UPDATE = 2048,\n"
"        PRIV_DELETE = 4096,\n"
"        PRIV_GRANT = 8192,\n"
"        PRIV_REVOKE = 16384,\n"
"        PRIV_ALL  = 4294967295\n"
"    };\n"
"\n"
"]]\n"
"\n"
"box.priv = {\n"
"    [\"R\"] = builtin.PRIV_R,\n"
"    [\"W\"] = builtin.PRIV_W,\n"
"    [\"X\"] = builtin.PRIV_X,\n"
"    [\"S\"] = builtin.PRIV_S,\n"
"    [\"U\"] = builtin.PRIV_U,\n"
"    [\"C\"] = builtin.PRIV_C,\n"
"    [\"D\"] = builtin.PRIV_D,\n"
"    [\"A\"] = builtin.PRIV_A,\n"
"    [\"REFERENCE\"] = builtin.PRIV_REFERENCE,\n"
"    [\"TRIGGER\"] = builtin.PRIV_TRIGGER,\n"
"    [\"INSERT\"] = builtin.PRIV_INSERT,\n"
"    [\"UPDATE\"] = builtin.PRIV_UPDATE,\n"
"    [\"DELETE\"] = builtin.PRIV_DELETE,\n"
"    [\"GRANT\"]= builtin.PRIV_GRANT,\n"
"    [\"REVOKE\"] = builtin.PRIV_REVOKE,\n"
"    [\"ALL\"] = builtin.PRIV_ALL\n"
"}\n"
"\n"
"local function user_or_role_resolve(user)\n"
"    local _vuser = box.space[box.schema.VUSER_ID]\n"
"    local tuple\n"
"    if type(user) == 'string' then\n"
"        tuple = _vuser.index.name:get{user}\n"
"    else\n"
"        tuple = _vuser:get{user}\n"
"    end\n"
"    if tuple == nil then\n"
"        return nil\n"
"    end\n"
"    return tuple[1]\n"
"end\n"
"\n"
"local function role_resolve(name_or_id)\n"
"    local _vuser = box.space[box.schema.VUSER_ID]\n"
"    local tuple\n"
"    if type(name_or_id) == 'string' then\n"
"        tuple = _vuser.index.name:get{name_or_id}\n"
"    elseif type(name_or_id) ~= 'nil' then\n"
"        tuple = _vuser:get{name_or_id}\n"
"    end\n"
"    if tuple == nil or tuple[4] ~= 'role' then\n"
"        return nil\n"
"    else\n"
"        return tuple[1]\n"
"    end\n"
"end\n"
"\n"
"local function user_resolve(name_or_id)\n"
"    local _vuser = box.space[box.schema.VUSER_ID]\n"
"    local tuple\n"
"    if type(name_or_id) == 'string' then\n"
"        tuple = _vuser.index.name:get{name_or_id}\n"
"    elseif type(name_or_id) ~= 'nil' then\n"
"        tuple = _vuser:get{name_or_id}\n"
"    end\n"
"    if tuple == nil or tuple[4] ~= 'user' then\n"
"        return nil\n"
"    else\n"
"        return tuple[1]\n"
"    end\n"
"end\n"
"\n"
"local function sequence_resolve(name_or_id)\n"
"    local _vsequence = box.space[box.schema.VSEQUENCE_ID]\n"
"    local tuple\n"
"    if type(name_or_id) == 'string' then\n"
"        tuple = _vsequence.index.name:get{name_or_id}\n"
"    elseif type(name_or_id) ~= 'nil' then\n"
"        tuple = _vsequence:get{name_or_id}\n"
"    end\n"
"    if tuple ~= nil then\n"
"        return tuple[1], tuple\n"
"    else\n"
"        return nil\n"
"    end\n"
"end\n"
"\n"
"-- Revoke all privileges associated with the given object.\n"
"local function revoke_object_privs(object_type, object_id)\n"
"    local _vpriv = box.space[box.schema.VPRIV_ID]\n"
"    local _priv = box.space[box.schema.PRIV_ID]\n"
"    local privs = _vpriv.index.object:select{object_type, object_id}\n"
"    for k, tuple in pairs(privs) do\n"
"        local uid = tuple[2]\n"
"        _priv:delete{uid, object_type, object_id}\n"
"    end\n"
"end\n"
"\n"
"-- Same as type(), but returns 'number' if 'param' is\n"
"-- of type 'cdata' and represents a 64-bit integer.\n"
"local function param_type(param)\n"
"    local t = type(param)\n"
"    if t == 'cdata' and tonumber64(param) ~= nil then\n"
"        t = 'number'\n"
"    end\n"
"    return t\n"
"end\n"
"\n"
"--[[\n"
" @brief Common function to check table with parameters (like options)\n"
" @param table - table with parameters\n"
" @param template - table with expected types of expected parameters\n"
"  type could be comma separated string with lua types (number, string etc),\n"
"  or 'any' if any type is allowed\n"
" The function checks following:\n"
" 1)that parameters table is a table (or nil)\n"
" 2)all keys in parameters are present in template\n"
" 3)type of every parameter fits (one of) types described in template\n"
" Check (2) and (3) could be disabled by adding {, dont_check = <smth is true>, }\n"
"  into parameters table\n"
" The functions calls box.error(box.error.ILLEGAL_PARAMS, ..) on error\n"
" @example\n"
" check_param_table(options, { user = 'string',\n"
"                              port = 'string, number',\n"
"                              data = 'any'} )\n"
"--]]\n"
"local function check_param_table(table, template)\n"
"    if table == nil then\n"
"        return\n"
"    end\n"
"    if type(table) ~= 'table' then\n"
"        box.error(box.error.ILLEGAL_PARAMS,\n"
"                  \"options should be a table\")\n"
"    end\n"
"    -- just pass {.. dont_check = true, ..} to disable checks below\n"
"    if table.dont_check then\n"
"        return\n"
"    end\n"
"    for k,v in pairs(table) do\n"
"        if template[k] == nil then\n"
"            box.error(box.error.ILLEGAL_PARAMS,\n"
"                      \"unexpected option '\" .. k .. \"'\")\n"
"        elseif template[k] == 'any' then\n"
"            -- any type is ok\n"
"        elseif (string.find(template[k], ',') == nil) then\n"
"            -- one type\n"
"            if param_type(v) ~= template[k] then\n"
"                box.error(box.error.ILLEGAL_PARAMS,\n"
"                          \"options parameter '\" .. k ..\n"
"                          \"' should be of type \" .. template[k])\n"
"            end\n"
"        else\n"
"            local good_types = string.gsub(template[k], ' ', '')\n"
"            local haystack = ',' .. good_types .. ','\n"
"            local needle = ',' .. param_type(v) .. ','\n"
"            if (string.find(haystack, needle) == nil) then\n"
"                good_types = string.gsub(good_types, ',', ', ')\n"
"                box.error(box.error.ILLEGAL_PARAMS,\n"
"                          \"options parameter '\" .. k ..\n"
"                          \"' should be one of types: \" .. template[k])\n"
"            end\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"--[[\n"
" @brief Common function to check type parameter (of function)\n"
" Calls box.error(box.error.ILLEGAL_PARAMS, ) on error\n"
" @example: check_param(user, 'user', 'string')\n"
"--]]\n"
"local function check_param(param, name, should_be_type)\n"
"    if param_type(param) ~= should_be_type then\n"
"        box.error(box.error.ILLEGAL_PARAMS,\n"
"                  name .. \" should be a \" .. should_be_type)\n"
"    end\n"
"end\n"
"\n"
"--[[\n"
" Adds to a table key-value pairs from defaults table\n"
"  that is not present in original table.\n"
" Returns updated table.\n"
" If nil is passed instead of table, it's treated as empty table {}\n"
" For example update_param_table({ type = 'hash', temporary = true },\n"
"                                { type = 'tree', unique = true })\n"
"  will return table { type = 'hash', temporary = true, unique = true }\n"
"--]]\n"
"local function update_param_table(table, defaults)\n"
"    local new_table = {}\n"
"    if defaults ~= nil then\n"
"        for k,v in pairs(defaults) do\n"
"            new_table[k] = v\n"
"        end\n"
"    end\n"
"    if table ~= nil then\n"
"        for k,v in pairs(table) do\n"
"            new_table[k] = v\n"
"        end\n"
"    end\n"
"    return new_table\n"
"end\n"
"\n"
"box.begin = function()\n"
"    if builtin.box_txn_begin() == -1 then\n"
"        box.error()\n"
"    end\n"
"end\n"
"\n"
"box.is_in_txn = builtin.box_txn\n"
"\n"
"box.savepoint = function()\n"
"    local csavepoint = builtin.box_txn_savepoint()\n"
"    if csavepoint == nil then\n"
"        box.error()\n"
"    end\n"
"    return { csavepoint=csavepoint, txn_id=builtin.box_txn_id() }\n"
"end\n"
"\n"
"local function atomic_tail(status, ...)\n"
"    if not status then\n"
"        box.rollback()\n"
"        error((...), 2)\n"
"     end\n"
"     box.commit()\n"
"     return ...\n"
"end\n"
"\n"
"box.atomic = function(fun, ...)\n"
"    box.begin()\n"
"    return atomic_tail(pcall(fun, ...))\n"
"end\n"
"\n"
"-- box.commit yields, so it's defined as Lua/C binding\n"
"-- box.rollback and box.rollback_to_savepoint yields as well\n"
"\n"
"local function update_format(format)\n"
"    local result = {}\n"
"    for i, given in ipairs(format) do\n"
"        local field = {}\n"
"        if type(given) ~= \"table\" then\n"
"            field.name = given\n"
"        else\n"
"            for k, v in pairs(given) do\n"
"                if k == 1 then\n"
"                    if given.name then\n"
"                        if not given.type then\n"
"                            field.type = v\n"
"                        else\n"
"                            field[1] = v\n"
"                        end\n"
"                    else\n"
"                        field.name = v\n"
"                    end\n"
"                elseif k == 2 and not given.type and not given.name then\n"
"                    field.type = v\n"
"                else\n"
"                    field[k] = v\n"
"                end\n"
"            end\n"
"        end\n"
"        if type(field.name) ~= 'string' then\n"
"            box.error(box.error.ILLEGAL_PARAMS,\n"
"                \"format[\" .. i .. \"]: name (string) is expected\")\n"
"        end\n"
"        if field.type == nil then\n"
"            field.type = 'any'\n"
"        elseif type(field.type) ~= 'string' then\n"
"            box.error(box.error.ILLEGAL_PARAMS,\n"
"                \"format[\" .. i .. \"]: type must be a string\")\n"
"        end\n"
"        table.insert(result, field)\n"
"    end\n"
"    return result\n"
"end\n"
"\n"
"box.schema.space = {}\n"
"box.schema.space.create = function(name, options)\n"
"    check_param(name, 'name', 'string')\n"
"    local options_template = {\n"
"        if_not_exists = 'boolean',\n"
"        engine = 'string',\n"
"        id = 'number',\n"
"        field_count = 'number',\n"
"        user = 'string, number',\n"
"        format = 'table',\n"
"        is_local = 'boolean',\n"
"        temporary = 'boolean',\n"
"    }\n"
"    local options_defaults = {\n"
"        engine = 'memtx',\n"
"        field_count = 0,\n"
"        temporary = false,\n"
"    }\n"
"    check_param_table(options, options_template)\n"
"    options = update_param_table(options, options_defaults)\n"
"\n"
"    local _space = box.space[box.schema.SPACE_ID]\n"
"    if box.space[name] then\n"
"        if options.if_not_exists then\n"
"            return box.space[name], \"not created\"\n"
"        else\n"
"            box.error(box.error.SPACE_EXISTS, name)\n"
"        end\n"
"    end\n"
"    local id = options.id\n"
"    if not id then\n"
"        local _schema = box.space._schema\n"
"        local max_id = _schema:update({'max_id'}, {{'+', 2, 1}})\n"
"        if max_id == nil then\n"
"            id = _space.index.primary:max()[1]\n"
"            if id < box.schema.SYSTEM_ID_MAX then\n"
"                id = box.schema.SYSTEM_ID_MAX\n"
"            end\n"
"            max_id = _schema:insert{'max_id', id + 1}\n"
"        end\n"
"        id = max_id[2]\n"
"    end\n"
"    local uid = session.euid()\n"
"    if options.user then\n"
"        uid = user_or_role_resolve(options.user)\n"
"        if uid == nil then\n"
"            box.error(box.error.NO_SUCH_USER, options.user)\n"
"        end\n"
"    end\n"
"    local format = options.format and options.format or {}\n"
"    check_param(format, 'format', 'table')\n"
"    format = update_format(format)\n"
"    -- filter out global parameters from the options array\n"
"    local space_options = setmap({\n"
"        group_id = options.is_local and 1 or nil,\n"
"        temporary = options.temporary and true or nil,\n"
"    })\n"
"    _space:insert{id, uid, name, options.engine, options.field_count,\n"
"        space_options, format}\n"
"    return box.space[id], \"created\"\n"
"end\n"
"\n"
"-- space format - the metadata about space fields\n"
"function box.schema.space.format(id, format)\n"
"    local _space = box.space._space\n"
"    local _vspace = box.space._vspace\n"
"    check_param(id, 'id', 'number')\n"
"\n"
"    if format == nil then\n"
"        local tuple = _vspace:get(id)\n"
"        if tuple == nil then\n"
"            box.error(box.error.NO_SUCH_SPACE, '#' .. tostring(id))\n"
"        end\n"
"        return tuple[7]\n"
"    else\n"
"        check_param(format, 'format', 'table')\n"
"        format = update_format(format)\n"
"        _space:update(id, {{'=', 7, format}})\n"
"    end\n"
"end\n"
"\n"
"box.schema.create_space = box.schema.space.create\n"
"\n"
"box.schema.space.drop = function(space_id, space_name, opts)\n"
"    check_param(space_id, 'space_id', 'number')\n"
"    opts = opts or {}\n"
"    check_param_table(opts, { if_exists = 'boolean' })\n"
"    local _space = box.space[box.schema.SPACE_ID]\n"
"    local _index = box.space[box.schema.INDEX_ID]\n"
"    local _vindex = box.space[box.schema.VINDEX_ID]\n"
"    local _truncate = box.space[box.schema.TRUNCATE_ID]\n"
"    local _space_sequence = box.space[box.schema.SPACE_SEQUENCE_ID]\n"
"    local sequence_tuple = _space_sequence:delete{space_id}\n"
"    if sequence_tuple ~= nil and sequence_tuple[3] == true then\n"
"        -- Delete automatically generated sequence.\n"
"        box.schema.sequence.drop(sequence_tuple[2])\n"
"    end\n"
"    local keys = _vindex:select(space_id)\n"
"    for i = #keys, 1, -1 do\n"
"        local v = keys[i]\n"
"        _index:delete{v[1], v[2]}\n"
"    end\n"
"    revoke_object_privs('space', space_id)\n"
"    _truncate:delete{space_id}\n"
"    if _space:delete{space_id} == nil then\n"
"        if space_name == nil then\n"
"            space_name = '#'..tostring(space_id)\n"
"        end\n"
"        if not opts.if_exists then\n"
"            box.error(box.error.NO_SUCH_SPACE, space_name)\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"box.schema.space.rename = function(space_id, space_name)\n"
"    check_param(space_id, 'space_id', 'number')\n"
"    check_param(space_name, 'space_name', 'string')\n"
"\n"
"    local _space = box.space[box.schema.SPACE_ID]\n"
"    _space:update(space_id, {{\"=\", 3, space_name}})\n"
"end\n"
"\n"
"box.schema.index = {}\n"
"\n"
"local function update_index_parts_1_6_0(parts)\n"
"    local result = {}\n"
"    if #parts % 2 ~= 0 then\n"
"        box.error(box.error.ILLEGAL_PARAMS,\n"
"                  \"options.parts: expected field_no (number), type (string) pairs\")\n"
"    end\n"
"    for i=1,#parts,2 do\n"
"        if type(parts[i]) ~= \"number\" then\n"
"            box.error(box.error.ILLEGAL_PARAMS,\n"
"                      \"options.parts: expected field_no (number), type (string) pairs\")\n"
"        elseif parts[i] == 0 then\n"
"            -- Lua uses one-based field numbers but _space is zero-based\n"
"            box.error(box.error.ILLEGAL_PARAMS,\n"
"                      \"invalid index parts: field_no must be one-based\")\n"
"        end\n"
"        if type(parts[i + 1]) ~= \"string\" then\n"
"            box.error(box.error.ILLEGAL_PARAMS,\n"
"                      \"options.parts: expected field_no (number), type (string) pairs\")\n"
"        end\n"
"        table.insert(result, {field = parts[i] - 1, type = parts[i + 1]})\n"
"    end\n"
"    return result\n"
"end\n"
"\n"
"local function update_index_parts(format, parts)\n"
"    if type(parts) ~= \"table\" then\n"
"        box.error(box.error.ILLEGAL_PARAMS,\n"
"        \"options.parts parameter should be a table\")\n"
"    end\n"
"    if #parts == 0 then\n"
"        box.error(box.error.ILLEGAL_PARAMS,\n"
"        \"options.parts must have at least one part\")\n"
"    end\n"
"    if type(parts[1]) == 'number' and type(parts[2]) == 'string' then\n"
"        return update_index_parts_1_6_0(parts), true\n"
"    end\n"
"\n"
"    local parts_can_be_simplified = true\n"
"    local result = {}\n"
"    for i=1,#parts do\n"
"        local part = {}\n"
"        if type(parts[i]) ~= \"table\" then\n"
"            part.field = parts[i]\n"
"        else\n"
"            for k, v in pairs(parts[i]) do\n"
"                -- Support {1, 'unsigned', collation='xx'} shortcut\n"
"                if k == 1 or k == 'field' then\n"
"                    part.field = v;\n"
"                elseif k == 2 or k == 'type' then\n"
"                    part.type = v;\n"
"                elseif k == 'collation' then\n"
"                    -- find ID by name\n"
"                    local coll = box.space._collation.index.name:get{v}\n"
"                    if not coll then\n"
"                        coll = box.space._collation.index.name:get{v:lower()}\n"
"                    end\n"
"                    if not coll then\n"
"                        box.error(box.error.ILLEGAL_PARAMS,\n"
"                            \"options.parts[\" .. i .. \"]: collation was not found by name '\" .. v .. \"'\")\n"
"                    end\n"
"                    part[k] = coll[1]\n"
"                    parts_can_be_simplified = false\n"
"                elseif k == 'is_nullable' then\n"
"                    part[k] = v\n"
"                    parts_can_be_simplified = false\n"
"                else\n"
"                    part[k] = v\n"
"                    parts_can_be_simplified = false\n"
"                end\n"
"            end\n"
"        end\n"
"        if type(part.field) ~= 'number' and type(part.field) ~= 'string' then\n"
"            box.error(box.error.ILLEGAL_PARAMS,\n"
"                      \"options.parts[\" .. i .. \"]: field (name or number) is expected\")\n"
"        elseif type(part.field) == 'string' then\n"
"            for k,v in pairs(format) do\n"
"                if v.name == part.field then\n"
"                    part.field = k\n"
"                    break\n"
"                end\n"
"            end\n"
"            if type(part.field) == 'string' then\n"
"                box.error(box.error.ILLEGAL_PARAMS,\n"
"                          \"options.parts[\" .. i .. \"]: field was not found by name '\" .. part.field .. \"'\")\n"
"            end\n"
"        elseif part.field == 0 then\n"
"            box.error(box.error.ILLEGAL_PARAMS,\n"
"                      \"options.parts[\" .. i .. \"]: field (number) must be one-based\")\n"
"        end\n"
"        local fmt = format[part.field]\n"
"        if part.type == nil then\n"
"            if fmt and fmt.type then\n"
"                part.type = fmt.type\n"
"            else\n"
"                part.type = 'scalar'\n"
"            end\n"
"        elseif type(part.type) ~= 'string' then\n"
"            box.error(box.error.ILLEGAL_PARAMS,\n"
"                      \"options.parts[\" .. i .. \"]: type (string) is expected\")\n"
"        end\n"
"        if part.is_nullable == nil then\n"
"            if fmt and fmt.is_nullable then\n"
"                part.is_nullable = true\n"
"                parts_can_be_simplified = false\n"
"            end\n"
"        elseif type(part.is_nullable) ~= 'boolean' then\n"
"            box.error(box.error.ILLEGAL_PARAMS,\n"
"                      \"options.parts[\" .. i .. \"]: type (boolean) is expected\")\n"
"        end\n"
"        part.field = part.field - 1\n"
"        table.insert(result, part)\n"
"    end\n"
"    return result, parts_can_be_simplified\n"
"end\n"
"\n"
"--\n"
"-- Convert index parts into 1.6.6 format if they\n"
"-- doesn't use collation and is_nullable options\n"
"--\n"
"local function simplify_index_parts(parts)\n"
"    local new_parts = {}\n"
"    for i, part in pairs(parts) do\n"
"        assert(part.collation == nil and part.is_nullable == nil,\n"
"               \"part is simple\")\n"
"        new_parts[i] = {part.field, part.type}\n"
"    end\n"
"    return new_parts\n"
"end\n"
"\n"
"-- Historically, some properties of an index\n"
"-- are stored as tuple fields, others in a\n"
"-- single field containing msgpack map.\n"
"-- This is the map.\n"
"local index_options = {\n"
"    unique = 'boolean',\n"
"    dimension = 'number',\n"
"    distance = 'string',\n"
"    run_count_per_level = 'number',\n"
"    run_size_ratio = 'number',\n"
"    range_size = 'number',\n"
"    page_size = 'number',\n"
"    bloom_fpr = 'number',\n"
"}\n"
"\n"
"--\n"
"-- check_param_table() template for alter index,\n"
"-- includes all index options.\n"
"--\n"
"local alter_index_template = {\n"
"    id = 'number',\n"
"    name = 'string',\n"
"    type = 'string',\n"
"    parts = 'table',\n"
"    sequence = 'boolean, number, string',\n"
"}\n"
"for k, v in pairs(index_options) do\n"
"    alter_index_template[k] = v\n"
"end\n"
"\n"
"--\n"
"-- check_param_table() template for create_index(), includes\n"
"-- all index options and if_not_exists specifier\n"
"--\n"
"local create_index_template = table.deepcopy(alter_index_template)\n"
"create_index_template.if_not_exists = \"boolean\"\n"
"\n"
"box.schema.index.create = function(space_id, name, options)\n"
"    check_param(space_id, 'space_id', 'number')\n"
"    check_param(name, 'name', 'string')\n"
"    check_param_table(options, create_index_template)\n"
"    local space = box.space[space_id]\n"
"    if not space then\n"
"        box.error(box.error.NO_SUCH_SPACE, '#'..tostring(space_id))\n"
"    end\n"
"    local format = space:format()\n"
"\n"
"    local options_defaults = {\n"
"        type = 'tree',\n"
"    }\n"
"    options = update_param_table(options, options_defaults)\n"
"    local type_dependent_defaults = {\n"
"        rtree = {parts = { 2, 'array' }, unique = false},\n"
"        bitset = {parts = { 2, 'unsigned' }, unique = false},\n"
"        other = {parts = { 1, 'unsigned' }, unique = true},\n"
"    }\n"
"    options_defaults = type_dependent_defaults[options.type]\n"
"            or type_dependent_defaults.other\n"
"    if not options.parts then\n"
"        local fieldno = options_defaults.parts[1]\n"
"        if #format >= fieldno then\n"
"            local t = format[fieldno].type\n"
"            if t ~= 'any' then\n"
"                options.parts = {{fieldno, format[fieldno].type}}\n"
"            end\n"
"        end\n"
"    end\n"
"    options = update_param_table(options, options_defaults)\n"
"    if space.engine == 'vinyl' then\n"
"        options_defaults = {\n"
"            page_size = box.cfg.vinyl_page_size,\n"
"            range_size = box.cfg.vinyl_range_size,\n"
"            run_count_per_level = box.cfg.vinyl_run_count_per_level,\n"
"            run_size_ratio = box.cfg.vinyl_run_size_ratio,\n"
"            bloom_fpr = box.cfg.vinyl_bloom_fpr\n"
"        }\n"
"    else\n"
"        options_defaults = {}\n"
"    end\n"
"    options = update_param_table(options, options_defaults)\n"
"\n"
"    local _index = box.space[box.schema.INDEX_ID]\n"
"    local _vindex = box.space[box.schema.VINDEX_ID]\n"
"    if _vindex.index.name:get{space_id, name} then\n"
"        if options.if_not_exists then\n"
"            return space.index[name], \"not created\"\n"
"        else\n"
"            box.error(box.error.INDEX_EXISTS, name)\n"
"        end\n"
"    end\n"
"\n"
"    local iid = 0\n"
"    if options.id then\n"
"        iid = options.id\n"
"    else\n"
"        -- max\n"
"        local tuple = _vindex.index[0]\n"
"            :select(space_id, { limit = 1, iterator = 'LE' })[1]\n"
"        if tuple then\n"
"            local id = tuple[1]\n"
"            if id == space_id then\n"
"                iid = tuple[2] + 1\n"
"            end\n"
"        end\n"
"    end\n"
"    local parts, parts_can_be_simplified =\n"
"        update_index_parts(format, options.parts)\n"
"    -- create_index() options contains type, parts, etc,\n"
"    -- stored separately. Remove these members from index_opts\n"
"    local index_opts = {\n"
"            dimension = options.dimension,\n"
"            unique = options.unique,\n"
"            distance = options.distance,\n"
"            page_size = options.page_size,\n"
"            range_size = options.range_size,\n"
"            run_count_per_level = options.run_count_per_level,\n"
"            run_size_ratio = options.run_size_ratio,\n"
"            bloom_fpr = options.bloom_fpr,\n"
"    }\n"
"    local field_type_aliases = {\n"
"        num = 'unsigned'; -- Deprecated since 1.7.2\n"
"        uint = 'unsigned';\n"
"        str = 'string';\n"
"        int = 'integer';\n"
"        ['*'] = 'any';\n"
"    };\n"
"    for _, part in pairs(parts) do\n"
"        local field_type = part.type:lower()\n"
"        part.type = field_type_aliases[field_type] or field_type\n"
"        if field_type == 'num' then\n"
"            log.warn(\"field type '%s' is deprecated since Tarantool 1.7, \"..\n"
"                     \"please use '%s' instead\", field_type, part.type)\n"
"        end\n"
"    end\n"
"    local _space_sequence = box.space[box.schema.SPACE_SEQUENCE_ID]\n"
"    local sequence_is_generated = false\n"
"    local sequence = options.sequence or nil -- ignore sequence = false\n"
"    if sequence ~= nil then\n"
"        if iid ~= 0 then\n"
"            box.error(box.error.MODIFY_INDEX, name, space.name,\n"
"                      \"sequence cannot be used with a secondary key\")\n"
"        end\n"
"        if #parts >= 1 and parts[1].type ~= 'integer' and\n"
"                           parts[1].type ~= 'unsigned' then\n"
"            box.error(box.error.MODIFY_INDEX, name, space.name,\n"
"                      \"sequence cannot be used with a non-integer key\")\n"
"        end\n"
"        if sequence == true then\n"
"            sequence = box.schema.sequence.create(space.name .. '_seq')\n"
"            sequence = sequence.id\n"
"            sequence_is_generated = true\n"
"        else\n"
"            sequence = sequence_resolve(sequence)\n"
"            if sequence == nil then\n"
"                box.error(box.error.NO_SUCH_SEQUENCE, options.sequence)\n"
"            end\n"
"        end\n"
"    end\n"
"    -- save parts in old format if possible\n"
"    if parts_can_be_simplified then\n"
"        parts = simplify_index_parts(parts)\n"
"    end\n"
"    _index:insert{space_id, iid, name, options.type, index_opts, parts}\n"
"    if sequence ~= nil then\n"
"        _space_sequence:insert{space_id, sequence, sequence_is_generated}\n"
"    end\n"
"    return space.index[name]\n"
"end\n"
"\n"
"box.schema.index.drop = function(space_id, index_id)\n"
"    check_param(space_id, 'space_id', 'number')\n"
"    check_param(index_id, 'index_id', 'number')\n"
"    if index_id == 0 then\n"
"        local _space_sequence = box.space[box.schema.SPACE_SEQUENCE_ID]\n"
"        local sequence_tuple = _space_sequence:delete{space_id}\n"
"        if sequence_tuple ~= nil and sequence_tuple[3] == true then\n"
"            -- Delete automatically generated sequence.\n"
"            box.schema.sequence.drop(sequence_tuple[2])\n"
"        end\n"
"    end\n"
"    local _index = box.space[box.schema.INDEX_ID]\n"
"    _index:delete{space_id, index_id}\n"
"end\n"
"\n"
"box.schema.index.rename = function(space_id, index_id, name)\n"
"    check_param(space_id, 'space_id', 'number')\n"
"    check_param(index_id, 'index_id', 'number')\n"
"    check_param(name, 'name', 'string')\n"
"\n"
"    local _index = box.space[box.schema.INDEX_ID]\n"
"    _index:update({space_id, index_id}, {{\"=\", 3, name}})\n"
"end\n"
"\n"
"box.schema.index.alter = function(space_id, index_id, options)\n"
"    local space = box.space[space_id]\n"
"    if space == nil then\n"
"        box.error(box.error.NO_SUCH_SPACE, '#'..tostring(space_id))\n"
"    end\n"
"    if space.index[index_id] == nil then\n"
"        box.error(box.error.NO_SUCH_INDEX, index_id, space.name)\n"
"    end\n"
"    if options == nil then\n"
"        return\n"
"    end\n"
"\n"
"    check_param_table(options, alter_index_template)\n"
"\n"
"    if type(space_id) ~= \"number\" then\n"
"        space_id = space.id\n"
"    end\n"
"    if type(index_id) ~= \"number\" then\n"
"        index_id = space.index[index_id].id\n"
"    end\n"
"    local format = space:format()\n"
"    local _index = box.space[box.schema.INDEX_ID]\n"
"    if options.id ~= nil then\n"
"        local can_update_field = {id = true, name = true, type = true }\n"
"        local can_update = true\n"
"        local cant_update_fields = ''\n"
"        for k,v in pairs(options) do\n"
"            if not can_update_field[k] then\n"
"                can_update = false\n"
"                cant_update_fields = cant_update_fields .. ' ' .. k\n"
"            end\n"
"        end\n"
"        if not can_update then\n"
"            box.error(box.error.PROC_LUA,\n"
"                      \"Don't know how to update both id and\" ..\n"
"                       cant_update_fields)\n"
"        end\n"
"        local ops = {}\n"
"        local function add_op(value, field_no)\n"
"            if value then\n"
"                table.insert(ops, {'=', field_no, value})\n"
"            end\n"
"        end\n"
"        add_op(options.id, 2)\n"
"        add_op(options.name, 3)\n"
"        add_op(options.type, 4)\n"
"        _index:update({space_id, index_id}, ops)\n"
"        return\n"
"    end\n"
"    local tuple = _index:get{space_id, index_id }\n"
"    local parts = {}\n"
"    local index_opts = {}\n"
"    local OPTS = 5\n"
"    local PARTS = 6\n"
"    if type(tuple[OPTS]) == 'number' then\n"
"        -- old format\n"
"        index_opts.unique = tuple[OPTS] == 1\n"
"        local part_count = tuple[PARTS]\n"
"        for i = 1, part_count do\n"
"            table.insert(parts, {tuple[2 * i + 4], tuple[2 * i + 5]});\n"
"        end\n"
"    else\n"
"        -- new format\n"
"        index_opts = tuple[OPTS]\n"
"        parts = tuple[PARTS]\n"
"    end\n"
"    if options.name == nil then\n"
"        options.name = tuple[3]\n"
"    end\n"
"    if options.type == nil then\n"
"        options.type = tuple[4]\n"
"    end\n"
"    for k, t in pairs(index_options) do\n"
"        if options[k] ~= nil then\n"
"            index_opts[k] = options[k]\n"
"        end\n"
"    end\n"
"    if options.parts then\n"
"        local parts_can_be_simplified\n"
"        parts, parts_can_be_simplified =\n"
"            update_index_parts(format, options.parts)\n"
"        -- save parts in old format if possible\n"
"        if parts_can_be_simplified then\n"
"            parts = simplify_index_parts(parts)\n"
"        end\n"
"    end\n"
"    local _space_sequence = box.space[box.schema.SPACE_SEQUENCE_ID]\n"
"    local sequence_is_generated = false\n"
"    local sequence = options.sequence\n"
"    local sequence_tuple\n"
"    if index_id ~= 0 then\n"
"        if sequence then\n"
"            box.error(box.error.MODIFY_INDEX, options.name, space.name,\n"
"                      \"sequence cannot be used with a secondary key\")\n"
"        end\n"
"        -- ignore 'sequence = false' for secondary indexes\n"
"        sequence = nil\n"
"    else\n"
"        sequence_tuple = _space_sequence:get(space_id)\n"
"        if (sequence or (sequence ~= false and sequence_tuple ~= nil)) and\n"
"           #parts >= 1 and (parts[1].type or parts[1][2]) ~= 'integer' and\n"
"                           (parts[1].type or parts[1][2]) ~= 'unsigned' then\n"
"            box.error(box.error.MODIFY_INDEX, options.name, space.name,\n"
"                      \"sequence cannot be used with a non-integer key\")\n"
"        end\n"
"    end\n"
"    if sequence == true then\n"
"        if sequence_tuple == nil or sequence_tuple[3] == false then\n"
"            sequence = box.schema.sequence.create(space.name .. '_seq')\n"
"            sequence = sequence.id\n"
"            sequence_is_generated = true\n"
"        else\n"
"            -- Space already has an automatically generated sequence.\n"
"            sequence = nil\n"
"        end\n"
"    elseif sequence then\n"
"        sequence = sequence_resolve(sequence)\n"
"        if sequence == nil then\n"
"            box.error(box.error.NO_SUCH_SEQUENCE, options.sequence)\n"
"        end\n"
"    end\n"
"    if sequence == false then\n"
"        _space_sequence:delete(space_id)\n"
"    end\n"
"    _index:replace{space_id, index_id, options.name, options.type,\n"
"                   index_opts, parts}\n"
"    if sequence then\n"
"        _space_sequence:replace{space_id, sequence, sequence_is_generated}\n"
"    end\n"
"    if sequence ~= nil and sequence_tuple ~= nil and\n"
"       sequence_tuple[3] == true and sequence_tuple[2] ~= sequence then\n"
"        -- Delete automatically generated sequence.\n"
"        box.schema.sequence.drop(sequence_tuple[2])\n"
"    end\n"
"end\n"
"\n"
"-- a static box_tuple_t ** instance for calling box_index_* API\n"
"local ptuple = ffi.new('box_tuple_t *[1]')\n"
"\n"
"local function keify(key)\n"
"    if key == nil then\n"
"        return {}\n"
"    elseif type(key) == \"table\" or is_tuple(key) then\n"
"        return key\n"
"    end\n"
"    return {key}\n"
"end\n"
"\n"
"local iterator_t = ffi.typeof('struct iterator')\n"
"ffi.metatype(iterator_t, {\n"
"    __tostring = function(iterator)\n"
"        return \"<iterator state>\"\n"
"    end;\n"
"})\n"
"\n"
"local iterator_gen = function(param, state)\n"
"    --[[\n"
"        index:pairs() mostly conforms to the Lua for-in loop conventions and\n"
"        tries to follow the best practices of Lua community.\n"
"\n"
"        - this generating function is stateless.\n"
"\n"
"        - *param* should contain **immutable** data needed to fully define\n"
"          an iterator. *param* is opaque for users. Currently it contains keybuf\n"
"          string just to prevent GC from collecting it. In future some other\n"
"          variables like space_id, index_id, sc_version will be stored here.\n"
"\n"
"        - *state* should contain **immutable** transient state of an iterator.\n"
"          *state* is opaque for users. Currently it contains `struct iterator`\n"
"          cdata that is modified during iteration. This is a sad limitation of\n"
"          underlying C API. Moreover, the separation of *param* and *state* is\n"
"          not properly implemented here. These drawbacks can be fixed in\n"
"          future without changing this API.\n"
"\n"
"        Please check out http://www.lua.org/pil/7.3.html for details.\n"
"    --]]\n"
"    if not ffi.istype(iterator_t, state) then\n"
"        error('usage: next(param, state)')\n"
"    end\n"
"    -- next() modifies state in-place\n"
"    if builtin.box_iterator_next(state, ptuple) ~= 0 then\n"
"        return box.error() -- error\n"
"    elseif ptuple[0] ~= nil then\n"
"        return state, tuple_bless(ptuple[0]) -- new state, value\n"
"    else\n"
"        return nil\n"
"    end\n"
"end\n"
"\n"
"local iterator_gen_luac = function(param, state)\n"
"    local tuple = internal.iterator_next(state)\n"
"    if tuple ~= nil then\n"
"        return state, tuple -- new state, value\n"
"    else\n"
"        return nil\n"
"    end\n"
"end\n"
"\n"
"-- global struct port instance to use by select()/get()\n"
"local port_tuple = ffi.new('struct port_tuple')\n"
"local port_tuple_entry_t = ffi.typeof('struct port_tuple_entry')\n"
"\n"
"-- Helper function to check space:method() usage\n"
"local function check_space_arg(space, method)\n"
"    if type(space) ~= 'table' or space.id == nil then\n"
"        local fmt = 'Use space:%s(...) instead of space.%s(...)'\n"
"        error(string.format(fmt, method, method))\n"
"    end\n"
"end\n"
"box.internal.check_space_arg = check_space_arg -- for net.box\n"
"\n"
"-- Helper function for nicer error messages\n"
"-- in some cases when space object is misused\n"
"-- Takes time so should not be used for DML.\n"
"local function check_space_exists(space)\n"
"    local s = box.space[space.id]\n"
"    if s == nil then\n"
"        box.error(box.error.NO_SUCH_SPACE, space.name)\n"
"    end\n"
"end\n"
"\n"
"-- Helper function to check index:method() usage\n"
"local function check_index_arg(index, method)\n"
"    if type(index) ~= 'table' or index.id == nil then\n"
"        local fmt = 'Use index:%s(...) instead of index.%s(...)'\n"
"        error(string.format(fmt, method, method))\n"
"    end\n"
"end\n"
"box.internal.check_index_arg = check_index_arg -- for net.box\n"
"\n"
"-- Helper function to check that space have primary key and return it\n"
"local function check_primary_index(space)\n"
"    local pk = space.index[0]\n"
"    if pk == nil then\n"
"        box.error(box.error.NO_SUCH_INDEX, 0, space.name)\n"
"    end\n"
"    return pk\n"
"end\n"
"box.internal.check_primary_index = check_primary_index -- for net.box\n"
"\n"
"box.internal.schema_version = builtin.box_schema_version\n"
"\n"
"local function check_iterator_type(opts, key_is_nil)\n"
"    local opts_type = type(opts)\n"
"    if opts ~= nil and opts_type ~= \"table\" and opts_type ~= \"string\" and opts_type ~= \"number\" then\n"
"        box.error(box.error.ITERATOR_TYPE, opts)\n"
"    end\n"
"\n"
"    local itype\n"
"    if opts_type == \"table\" and opts.iterator then\n"
"        if type(opts.iterator) == \"number\" then\n"
"            itype = opts.iterator\n"
"        elseif type(opts.iterator) == \"string\" then\n"
"            itype = box.index[string.upper(opts.iterator)]\n"
"            if itype == nil then\n"
"                box.error(box.error.ITERATOR_TYPE, opts.iterator)\n"
"            end\n"
"        else\n"
"            box.error(box.error.ITERATOR_TYPE, tostring(opts.iterator))\n"
"        end\n"
"    elseif opts_type == \"number\" then\n"
"        itype = opts\n"
"    elseif opts_type == \"string\" then\n"
"        itype = box.index[string.upper(opts)]\n"
"        if itype == nil then\n"
"            box.error(box.error.ITERATOR_TYPE, opts)\n"
"        end\n"
"    else\n"
"        -- Use ALL for {} and nil keys and EQ for other keys\n"
"        itype = key_is_nil and box.index.ALL or box.index.EQ\n"
"    end\n"
"    return itype\n"
"end\n"
"\n"
"internal.check_iterator_type = check_iterator_type\n"
"\n"
"local base_index_mt = {}\n"
"base_index_mt.__index = base_index_mt\n"
"--\n"
"-- Inherit engine specific index metatables from a base one.\n"
"--\n"
"local vinyl_index_mt = {}\n"
"vinyl_index_mt.__index = vinyl_index_mt\n"
"local memtx_index_mt = {}\n"
"memtx_index_mt.__index = memtx_index_mt\n"
"--\n"
"-- When a new method is added below to base index mt, the same\n"
"-- method is added both to vinyl and memtx index mt.\n"
"--\n"
"setmetatable(base_index_mt, {\n"
"    __newindex = function(t, k, v)\n"
"        vinyl_index_mt[k] = v\n"
"        memtx_index_mt[k] = v\n"
"        rawset(t, k, v)\n"
"    end\n"
"})\n"
"-- __len and __index\n"
"base_index_mt.len = function(index)\n"
"    check_index_arg(index, 'len')\n"
"    local ret = builtin.box_index_len(index.space_id, index.id)\n"
"    if ret == -1 then\n"
"        box.error()\n"
"    end\n"
"    return tonumber(ret)\n"
"end\n"
"-- index.bsize\n"
"base_index_mt.bsize = function(index)\n"
"    check_index_arg(index, 'bsize')\n"
"    local ret = builtin.box_index_bsize(index.space_id, index.id)\n"
"    if ret == -1 then\n"
"        box.error()\n"
"    end\n"
"    return tonumber(ret)\n"
"end\n"
"-- Lua 5.2 compatibility\n"
"base_index_mt.__len = base_index_mt.len\n"
"-- min and max\n"
"base_index_mt.min_ffi = function(index, key)\n"
"    check_index_arg(index, 'min')\n"
"    local ibuf = cord_ibuf_take()\n"
"    local pkey, pkey_end = tuple_encode(ibuf, key)\n"
"    local nok = builtin.box_index_min(index.space_id, index.id, pkey, pkey_end,\n"
"                                      ptuple) ~= 0\n"
"    cord_ibuf_put(ibuf)\n"
"    if nok then\n"
"        box.error() -- error\n"
"    elseif ptuple[0] ~= nil then\n"
"        return tuple_bless(ptuple[0])\n"
"    else\n"
"        return\n"
"    end\n"
"end\n"
"base_index_mt.min_luac = function(index, key)\n"
"    check_index_arg(index, 'min')\n"
"    key = keify(key)\n"
"    return internal.min(index.space_id, index.id, key);\n"
"end\n"
"base_index_mt.max_ffi = function(index, key)\n"
"    check_index_arg(index, 'max')\n"
"    local ibuf = cord_ibuf_take()\n"
"    local pkey, pkey_end = tuple_encode(ibuf, key)\n"
"    local nok = builtin.box_index_max(index.space_id, index.id, pkey, pkey_end,\n"
"                                      ptuple) ~= 0\n"
"    cord_ibuf_put(ibuf)\n"
"    if nok then\n"
"        box.error() -- error\n"
"    elseif ptuple[0] ~= nil then\n"
"        return tuple_bless(ptuple[0])\n"
"    else\n"
"        return\n"
"    end\n"
"end\n"
"base_index_mt.max_luac = function(index, key)\n"
"    check_index_arg(index, 'max')\n"
"    key = keify(key)\n"
"    return internal.max(index.space_id, index.id, key);\n"
"end\n"
"base_index_mt.random_ffi = function(index, rnd)\n"
"    check_index_arg(index, 'random')\n"
"    rnd = rnd or math.random()\n"
"    if builtin.box_index_random(index.space_id, index.id, rnd,\n"
"                                ptuple) ~= 0 then\n"
"        box.error() -- error\n"
"    elseif ptuple[0] ~= nil then\n"
"        return tuple_bless(ptuple[0])\n"
"    else\n"
"        return\n"
"    end\n"
"end\n"
"base_index_mt.random_luac = function(index, rnd)\n"
"    check_index_arg(index, 'random')\n"
"    rnd = rnd or math.random()\n"
"    return internal.random(index.space_id, index.id, rnd);\n"
"end\n"
"-- iteration\n"
"base_index_mt.pairs_ffi = function(index, key, opts)\n"
"    check_index_arg(index, 'pairs')\n"
"    local ibuf = cord_ibuf_take()\n"
"    local pkey, pkey_end = tuple_encode(ibuf, key)\n"
"    local itype = check_iterator_type(opts, pkey + 1 >= pkey_end);\n"
"\n"
"    local keybuf = ffi.string(pkey, pkey_end - pkey)\n"
"    cord_ibuf_put(ibuf)\n"
"    local pkeybuf = ffi.cast('const char *', keybuf)\n"
"    local cdata = builtin.box_index_iterator(index.space_id, index.id,\n"
"        itype, pkeybuf, pkeybuf + #keybuf);\n"
"    if cdata == nil then\n"
"        box.error()\n"
"    end\n"
"    return fun.wrap(iterator_gen, keybuf,\n"
"        ffi.gc(cdata, builtin.box_iterator_free))\n"
"end\n"
"base_index_mt.pairs_luac = function(index, key, opts)\n"
"    check_index_arg(index, 'pairs')\n"
"    key = keify(key)\n"
"    local itype = check_iterator_type(opts, #key == 0);\n"
"    local keymp = msgpack.encode(key)\n"
"    local keybuf = ffi.string(keymp, #keymp)\n"
"    local cdata = internal.iterator(index.space_id, index.id, itype, keymp);\n"
"    return fun.wrap(iterator_gen_luac, keybuf,\n"
"        ffi.gc(cdata, builtin.box_iterator_free))\n"
"end\n"
"\n"
"-- index subtree size\n"
"base_index_mt.count_ffi = function(index, key, opts)\n"
"    check_index_arg(index, 'count')\n"
"    local ibuf = cord_ibuf_take()\n"
"    local pkey, pkey_end = tuple_encode(ibuf, key)\n"
"    local itype = check_iterator_type(opts, pkey + 1 >= pkey_end);\n"
"    local count = builtin.box_index_count(index.space_id, index.id,\n"
"        itype, pkey, pkey_end);\n"
"    cord_ibuf_put(ibuf)\n"
"    if count == -1 then\n"
"        box.error()\n"
"    end\n"
"    return tonumber(count)\n"
"end\n"
"base_index_mt.count_luac = function(index, key, opts)\n"
"    check_index_arg(index, 'count')\n"
"    key = keify(key)\n"
"    local itype = check_iterator_type(opts, #key == 0);\n"
"    return internal.count(index.space_id, index.id, itype, key);\n"
"end\n"
"\n"
"base_index_mt.get_ffi = function(index, key)\n"
"    check_index_arg(index, 'get')\n"
"    local ibuf = cord_ibuf_take()\n"
"    local key, key_end = tuple_encode(ibuf, key)\n"
"    local nok = builtin.box_index_get(index.space_id, index.id, key, key_end,\n"
"                                      ptuple) ~= 0\n"
"    cord_ibuf_put(ibuf)\n"
"    if nok then\n"
"        return box.error() -- error\n"
"    elseif ptuple[0] ~= nil then\n"
"        return tuple_bless(ptuple[0])\n"
"    else\n"
"        return\n"
"    end\n"
"end\n"
"base_index_mt.get_luac = function(index, key)\n"
"    check_index_arg(index, 'get')\n"
"    key = keify(key)\n"
"    return internal.get(index.space_id, index.id, key)\n"
"end\n"
"\n"
"local function check_select_opts(opts, key_is_nil)\n"
"    local offset = 0\n"
"    local limit = 4294967295\n"
"    local iterator = check_iterator_type(opts, key_is_nil)\n"
"    if opts ~= nil and type(opts) == \"table\" then\n"
"        if opts.offset ~= nil then\n"
"            offset = opts.offset\n"
"        end\n"
"        if opts.limit ~= nil then\n"
"            limit = opts.limit\n"
"        end\n"
"    end\n"
"    return iterator, offset, limit\n"
"end\n"
"\n"
"box.internal.check_select_opts = check_select_opts -- for net.box\n"
"\n"
"base_index_mt.select_ffi = function(index, key, opts)\n"
"    check_index_arg(index, 'select')\n"
"    local ibuf = cord_ibuf_take()\n"
"    local key, key_end = tuple_encode(ibuf, key)\n"
"    local iterator, offset, limit = check_select_opts(opts, key + 1 >= key_end)\n"
"\n"
"    local port = ffi.cast('struct port *', port_tuple)\n"
"\n"
"    local nok = builtin.box_select(index.space_id, index.id, iterator, offset,\n"
"                                   limit, key, key_end, port) ~= 0\n"
"    cord_ibuf_put(ibuf)\n"
"    if nok then\n"
"        return box.error()\n"
"    end\n"
"\n"
"    local ret = {}\n"
"    local entry = port_tuple.first\n"
"    for i=1,tonumber(port_tuple.size),1 do\n"
"        ret[i] = tuple_bless(entry.tuple)\n"
"        entry = entry.next\n"
"    end\n"
"    builtin.port_destroy(port);\n"
"    return ret\n"
"end\n"
"\n"
"base_index_mt.select_luac = function(index, key, opts)\n"
"    check_index_arg(index, 'select')\n"
"    local key = keify(key)\n"
"    local iterator, offset, limit = check_select_opts(opts, #key == 0)\n"
"    return internal.select(index.space_id, index.id, iterator,\n"
"        offset, limit, key)\n"
"end\n"
"\n"
"base_index_mt.update = function(index, key, ops)\n"
"    check_index_arg(index, 'update')\n"
"    return internal.update(index.space_id, index.id, keify(key), ops);\n"
"end\n"
"base_index_mt.delete = function(index, key)\n"
"    check_index_arg(index, 'delete')\n"
"    return internal.delete(index.space_id, index.id, keify(key));\n"
"end\n"
"\n"
"base_index_mt.stat = function(index)\n"
"    return internal.stat(index.space_id, index.id);\n"
"end\n"
"\n"
"base_index_mt.compact = function(index)\n"
"    return internal.compact(index.space_id, index.id)\n"
"end\n"
"\n"
"base_index_mt.drop = function(index)\n"
"    check_index_arg(index, 'drop')\n"
"    return box.schema.index.drop(index.space_id, index.id)\n"
"end\n"
"base_index_mt.rename = function(index, name)\n"
"    check_index_arg(index, 'rename')\n"
"    return box.schema.index.rename(index.space_id, index.id, name)\n"
"end\n"
"base_index_mt.alter = function(index, options)\n"
"    check_index_arg(index, 'alter')\n"
"    if index.id == nil or index.space_id == nil then\n"
"        box.error(box.error.PROC_LUA, \"Usage: index:alter{opts}\")\n"
"    end\n"
"    return box.schema.index.alter(index.space_id, index.id, options)\n"
"end\n"
"\n"
"local read_ops = {'select', 'get', 'min', 'max', 'count', 'random', 'pairs'}\n"
"for _, op in ipairs(read_ops) do\n"
"    vinyl_index_mt[op] = base_index_mt[op..'_luac']\n"
"    memtx_index_mt[op] = base_index_mt[op..'_ffi']\n"
"end\n"
"-- Lua 5.2 compatibility\n"
"vinyl_index_mt.__pairs = vinyl_index_mt.pairs\n"
"vinyl_index_mt.__ipairs = vinyl_index_mt.pairs\n"
"memtx_index_mt.__pairs = memtx_index_mt.pairs\n"
"memtx_index_mt.__ipairs = memtx_index_mt.pairs\n"
"\n"
"local space_mt = {}\n"
"space_mt.len = function(space)\n"
"    check_space_arg(space, 'len')\n"
"    local pk = space.index[0]\n"
"    if pk == nil then\n"
"        return 0 -- empty space without indexes, return 0\n"
"    end\n"
"    return space.index[0]:len()\n"
"end\n"
"space_mt.count = function(space, key, opts)\n"
"    check_space_arg(space, 'count')\n"
"    local pk = space.index[0]\n"
"    if pk == nil then\n"
"        return 0 -- empty space without indexes, return 0\n"
"    end\n"
"    return pk:count(key, opts)\n"
"end\n"
"space_mt.bsize = function(space)\n"
"    check_space_arg(space, 'bsize')\n"
"    local s = builtin.space_by_id(space.id)\n"
"    if s == nil then\n"
"        box.error(box.error.NO_SUCH_SPACE, space.name)\n"
"    end\n"
"    return builtin.space_bsize(s)\n"
"end\n"
"\n"
"space_mt.get = function(space, key)\n"
"    check_space_arg(space, 'get')\n"
"    return check_primary_index(space):get(key)\n"
"end\n"
"space_mt.select = function(space, key, opts)\n"
"    check_space_arg(space, 'select')\n"
"    return check_primary_index(space):select(key, opts)\n"
"end\n"
"space_mt.insert = function(space, tuple)\n"
"    check_space_arg(space, 'insert')\n"
"    return internal.insert(space.id, tuple);\n"
"end\n"
"space_mt.replace = function(space, tuple)\n"
"    check_space_arg(space, 'replace')\n"
"    return internal.replace(space.id, tuple);\n"
"end\n"
"space_mt.put = space_mt.replace; -- put is an alias for replace\n"
"space_mt.update = function(space, key, ops)\n"
"    check_space_arg(space, 'update')\n"
"    return check_primary_index(space):update(key, ops)\n"
"end\n"
"space_mt.upsert = function(space, tuple_key, ops, deprecated)\n"
"    check_space_arg(space, 'upsert')\n"
"    if deprecated ~= nil then\n"
"        local msg = \"Error: extra argument in upsert call: \"\n"
"        msg = msg .. tostring(deprecated)\n"
"        msg = msg .. \". Usage :upsert(tuple, operations)\"\n"
"        box.error(box.error.PROC_LUA, msg)\n"
"    end\n"
"    return internal.upsert(space.id, tuple_key, ops);\n"
"end\n"
"space_mt.delete = function(space, key)\n"
"    check_space_arg(space, 'delete')\n"
"    return check_primary_index(space):delete(key)\n"
"end\n"
"-- Assumes that spaceno has a TREE (NUM) primary key\n"
"-- inserts a tuple after getting the next value of the\n"
"-- primary key and returns it back to the user\n"
"space_mt.auto_increment = function(space, tuple)\n"
"    check_space_arg(space, 'auto_increment')\n"
"    local max_tuple = check_primary_index(space):max()\n"
"    local max = 0\n"
"    if max_tuple ~= nil then\n"
"        max = max_tuple[1]\n"
"    end\n"
"    table.insert(tuple, 1, max + 1)\n"
"    return space:insert(tuple)\n"
"end\n"
"\n"
"space_mt.pairs = function(space, key, opts)\n"
"    check_space_arg(space, 'pairs')\n"
"    local pk = space.index[0]\n"
"    if pk == nil then\n"
"        -- empty space without indexes, return empty iterator\n"
"        return fun.iter({})\n"
"    end\n"
"    return pk:pairs(key, opts)\n"
"end\n"
"space_mt.__pairs = space_mt.pairs -- Lua 5.2 compatibility\n"
"space_mt.__ipairs = space_mt.pairs -- Lua 5.2 compatibility\n"
"space_mt.truncate = function(space)\n"
"    check_space_arg(space, 'truncate')\n"
"    return internal.truncate(space.id)\n"
"end\n"
"space_mt.format = function(space, format)\n"
"    check_space_arg(space, 'format')\n"
"    return box.schema.space.format(space.id, format)\n"
"end\n"
"space_mt.drop = function(space)\n"
"    check_space_arg(space, 'drop')\n"
"    check_space_exists(space)\n"
"    return box.schema.space.drop(space.id, space.name)\n"
"end\n"
"space_mt.rename = function(space, name)\n"
"    check_space_arg(space, 'rename')\n"
"    check_space_exists(space)\n"
"    return box.schema.space.rename(space.id, name)\n"
"end\n"
"space_mt.create_index = function(space, name, options)\n"
"    check_space_arg(space, 'create_index')\n"
"    check_space_exists(space)\n"
"    return box.schema.index.create(space.id, name, options)\n"
"end\n"
"space_mt.run_triggers = function(space, yesno)\n"
"    check_space_arg(space, 'run_triggers')\n"
"    local s = builtin.space_by_id(space.id)\n"
"    if s == nil then\n"
"        box.error(box.error.NO_SUCH_SPACE, space.name)\n"
"    end\n"
"    builtin.space_run_triggers(s, yesno)\n"
"end\n"
"space_mt.frommap = box.internal.space.frommap\n"
"space_mt.__index = space_mt\n"
"\n"
"box.schema.index_mt = base_index_mt\n"
"box.schema.memtx_index_mt = memtx_index_mt\n"
"box.schema.vinyl_index_mt = vinyl_index_mt\n"
"box.schema.space_mt = space_mt\n"
"\n"
"--\n"
"-- Wrap a global space/index metatable into a space/index local\n"
"-- one. Routinely this metatable just indexes the global one. When\n"
"-- a user attempts to extend a space or index methods via local\n"
"-- space/index metatable instead of from box.schema mt, the local\n"
"-- metatable is transformed. Its __index metamethod starts looking\n"
"-- up at first in self, and only then into the global mt.\n"
"--\n"
"local function wrap_schema_object_mt(name)\n"
"    local global_mt = box.schema[name]\n"
"    local mt = {\n"
"        __index = global_mt,\n"
"        __ipairs = global_mt.__ipairs,\n"
"        __pairs = global_mt.__pairs\n"
"    }\n"
"    local mt_mt = {}\n"
"    mt_mt.__newindex = function(t, k, v)\n"
"        mt_mt.__newindex = nil\n"
"        mt.__index = function(t, k)\n"
"            return mt[k] or box.schema[name][k]\n"
"        end\n"
"        rawset(mt, k, v)\n"
"    end\n"
"    setmetatable(mt, mt_mt)\n"
"    return mt\n"
"end\n"
"\n"
"function box.schema.space.bless(space)\n"
"    local index_mt_name\n"
"    if space.engine == 'vinyl' then\n"
"        index_mt_name = 'vinyl_index_mt'\n"
"    else\n"
"        index_mt_name = 'memtx_index_mt'\n"
"    end\n"
"    local space_mt = wrap_schema_object_mt('space_mt')\n"
"\n"
"    setmetatable(space, space_mt)\n"
"    if type(space.index) == 'table' and space.enabled then\n"
"        for j, index in pairs(space.index) do\n"
"            if type(j) == 'number' then\n"
"                setmetatable(index, wrap_schema_object_mt(index_mt_name))\n"
"            end\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"local sequence_mt = {}\n"
"sequence_mt.__index = sequence_mt\n"
"\n"
"sequence_mt.next = function(self)\n"
"    return internal.sequence.next(self.id)\n"
"end\n"
"\n"
"sequence_mt.set = function(self, value)\n"
"    return internal.sequence.set(self.id, value)\n"
"end\n"
"\n"
"sequence_mt.reset = function(self)\n"
"    return internal.sequence.reset(self.id)\n"
"end\n"
"\n"
"sequence_mt.alter = function(self, opts)\n"
"    box.schema.sequence.alter(self.id, opts)\n"
"end\n"
"\n"
"sequence_mt.drop = function(self)\n"
"    box.schema.sequence.drop(self.id)\n"
"end\n"
"\n"
"local function sequence_tuple_decode(seq, tuple)\n"
"    seq.id, seq.uid, seq.name, seq.step, seq.min, seq.max,\n"
"        seq.start, seq.cache, seq.cycle = tuple:unpack()\n"
"end\n"
"\n"
"local function sequence_new(tuple)\n"
"    local seq = setmetatable({}, sequence_mt)\n"
"    sequence_tuple_decode(seq, tuple)\n"
"    return seq\n"
"end\n"
"\n"
"local function sequence_on_alter(old_tuple, new_tuple)\n"
"    if old_tuple and not new_tuple then\n"
"        local old_name = old_tuple[3]\n"
"        box.sequence[old_name] = nil\n"
"    elseif not old_tuple and new_tuple then\n"
"        local seq = sequence_new(new_tuple)\n"
"        box.sequence[seq.name] = seq\n"
"    else\n"
"        local old_name = old_tuple[3]\n"
"        local seq = box.sequence[old_name]\n"
"        if not seq then\n"
"            seq = sequence_new(seq, new_tuple)\n"
"        else\n"
"            sequence_tuple_decode(seq, new_tuple)\n"
"        end\n"
"        box.sequence[old_name] = nil\n"
"        box.sequence[seq.name] = seq\n"
"    end\n"
"end\n"
"\n"
"box.sequence = {}\n"
"local function box_sequence_init()\n"
"    -- Install a trigger that will update Lua objects on\n"
"    -- _sequence space modifications.\n"
"    internal.sequence.on_alter(sequence_on_alter)\n"
"end\n"
"\n"
"local sequence_options = {\n"
"    step = 'number',\n"
"    min = 'number',\n"
"    max = 'number',\n"
"    start = 'number',\n"
"    cache = 'number',\n"
"    cycle = 'boolean',\n"
"}\n"
"\n"
"local create_sequence_options = table.deepcopy(sequence_options)\n"
"create_sequence_options.if_not_exists = 'boolean'\n"
"\n"
"local alter_sequence_options = table.deepcopy(sequence_options)\n"
"alter_sequence_options.name = 'string'\n"
"\n"
"box.schema.sequence = {}\n"
"box.schema.sequence.create = function(name, opts)\n"
"    opts = opts or {}\n"
"    check_param(name, 'name', 'string')\n"
"    check_param_table(opts, create_sequence_options)\n"
"    local ascending = not opts.step or opts.step > 0\n"
"    local options_defaults = {\n"
"        step = 1,\n"
"        min = ascending and 1 or INT64_MIN,\n"
"        max = ascending and INT64_MAX or -1,\n"
"        start = ascending and (opts.min or 1) or (opts.max or -1),\n"
"        cache = 0,\n"
"        cycle = false,\n"
"    }\n"
"    opts = update_param_table(opts, options_defaults)\n"
"    local id = sequence_resolve(name)\n"
"    if id ~= nil then\n"
"        if not opts.if_not_exists then\n"
"            box.error(box.error.SEQUENCE_EXISTS, name)\n"
"        end\n"
"        return box.sequence[name], 'not created'\n"
"    end\n"
"    local _sequence = box.space[box.schema.SEQUENCE_ID]\n"
"    _sequence:auto_increment{session.euid(), name, opts.step, opts.min,\n"
"                             opts.max, opts.start, opts.cache, opts.cycle}\n"
"    return box.sequence[name]\n"
"end\n"
"\n"
"box.schema.sequence.alter = function(name, opts)\n"
"    check_param_table(opts, alter_sequence_options)\n"
"    local id, tuple = sequence_resolve(name)\n"
"    if id == nil then\n"
"        box.error(box.error.NO_SUCH_SEQUENCE, name)\n"
"    end\n"
"    if opts == nil then\n"
"        return\n"
"    end\n"
"    local seq = {}\n"
"    sequence_tuple_decode(seq, tuple)\n"
"    opts = update_param_table(opts, seq)\n"
"    local _sequence = box.space[box.schema.SEQUENCE_ID]\n"
"    _sequence:replace{seq.id, seq.uid, opts.name, opts.step, opts.min,\n"
"                      opts.max, opts.start, opts.cache, opts.cycle}\n"
"end\n"
"\n"
"box.schema.sequence.drop = function(name, opts)\n"
"    opts = opts or {}\n"
"    check_param_table(opts, {if_exists = 'boolean'})\n"
"    local id = sequence_resolve(name)\n"
"    if id == nil then\n"
"        if not opts.if_exists then\n"
"            box.error(box.error.NO_SUCH_SEQUENCE, name)\n"
"        end\n"
"        return\n"
"    end\n"
"    revoke_object_privs('sequence', id)\n"
"    local _sequence = box.space[box.schema.SEQUENCE_ID]\n"
"    local _sequence_data = box.space[box.schema.SEQUENCE_DATA_ID]\n"
"    _sequence_data:delete{id}\n"
"    _sequence:delete{id}\n"
"end\n"
"\n"
"local function privilege_resolve(privilege)\n"
"    local numeric = 0\n"
"    if type(privilege) == 'string' then\n"
"        privilege = string.lower(privilege)\n"
"        if string.find(privilege, 'read') then\n"
"            numeric = numeric + box.priv.R\n"
"        end\n"
"        if string.find(privilege, 'write') then\n"
"            numeric = numeric + box.priv.W\n"
"        end\n"
"        if string.find(privilege, 'execute') then\n"
"            numeric = numeric + box.priv.X\n"
"        end\n"
"        if string.find(privilege, 'session') then\n"
"            numeric = numeric + box.priv.S\n"
"        end\n"
"        if string.find(privilege, 'usage') then\n"
"            numeric = numeric + box.priv.U\n"
"        end\n"
"        if string.find(privilege, 'create') then\n"
"            numeric = numeric + box.priv.C\n"
"        end\n"
"        if string.find(privilege, 'drop') then\n"
"            numeric = numeric + box.priv.D\n"
"        end\n"
"        if string.find(privilege, 'alter') then\n"
"            numeric = numeric + box.priv.A\n"
"        end\n"
"        if string.find(privilege, 'reference') then\n"
"            numeric = numeric + box.priv.REFERENCE\n"
"        end\n"
"        if string.find(privilege, 'trigger') then\n"
"            numeric = numeric + box.priv.TRIGGER\n"
"        end\n"
"        if string.find(privilege, 'insert') then\n"
"            numeric = numeric + box.priv.INSERT\n"
"        end\n"
"        if string.find(privilege, 'update') then\n"
"            numeric = numeric + box.priv.UPDATE\n"
"        end\n"
"        if string.find(privilege, 'delete') then\n"
"            numeric = numeric + box.priv.DELETE\n"
"        end\n"
"    else\n"
"        numeric = privilege\n"
"    end\n"
"    return numeric\n"
"end\n"
"\n"
"-- allowed combination of privilege bits for object\n"
"local priv_object_combo = {\n"
"    [\"universe\"] = box.priv.ALL,\n"
"-- sic: we allow to grant 'execute' on space. This is a legacy\n"
"-- bug, please fix it in 2.0\n"
"    [\"space\"]    = bit.bxor(box.priv.ALL, box.priv.S,\n"
"                            box.priv.REVOKE, box.priv.GRANT),\n"
"    [\"sequence\"] = bit.bor(box.priv.R, box.priv.W, box.priv.U,\n"
"                           box.priv.C, box.priv.A, box.priv.D),\n"
"    [\"function\"] = bit.bor(box.priv.X, box.priv.U,\n"
"                           box.priv.C, box.priv.D),\n"
"    [\"role\"]     = bit.bor(box.priv.X, box.priv.U,\n"
"                           box.priv.C, box.priv.D),\n"
"    [\"user\"]     = bit.bor(box.priv.C, box.priv.A,\n"
"                           box.priv.D),\n"
"}\n"
"\n"
"--\n"
"-- Resolve privilege hex by name and check\n"
"-- that bits are allowed for this object type\n"
"--\n"
"local function privilege_check(privilege, object_type)\n"
"    local priv_hex = privilege_resolve(privilege)\n"
"    if priv_object_combo[object_type] == nil then\n"
"        box.error(box.error.UNKNOWN_SCHEMA_OBJECT, object_type)\n"
"    elseif type(priv_hex) ~= 'number' or priv_hex == 0 or\n"
"           bit.band(priv_hex, priv_object_combo[object_type] or 0) ~= priv_hex then\n"
"        box.error(box.error.UNSUPPORTED_PRIV, object_type, privilege)\n"
"    end\n"
"    return priv_hex\n"
"end\n"
"\n"
"local function privilege_name(privilege)\n"
"    local names = {}\n"
"    if bit.band(privilege, box.priv.R) ~= 0 then\n"
"        table.insert(names, \"read\")\n"
"    end\n"
"    if bit.band(privilege, box.priv.W) ~= 0 then\n"
"        table.insert(names, \"write\")\n"
"    end\n"
"    if bit.band(privilege, box.priv.X) ~= 0 then\n"
"        table.insert(names, \"execute\")\n"
"    end\n"
"    if bit.band(privilege, box.priv.S) ~= 0 then\n"
"        table.insert(names, \"session\")\n"
"    end\n"
"    if bit.band(privilege, box.priv.U) ~= 0 then\n"
"        table.insert(names, \"usage\")\n"
"    end\n"
"    if bit.band(privilege, box.priv.C) ~= 0 then\n"
"        table.insert(names, \"create\")\n"
"    end\n"
"    if bit.band(privilege, box.priv.D) ~= 0 then\n"
"        table.insert(names, \"drop\")\n"
"    end\n"
"    if bit.band(privilege, box.priv.A) ~= 0 then\n"
"        table.insert(names, \"alter\")\n"
"    end\n"
"    if bit.band(privilege, box.priv.REFERENCE) ~= 0 then\n"
"        table.insert(names, \"reference\")\n"
"    end\n"
"    if bit.band(privilege, box.priv.TRIGGER) ~= 0 then\n"
"        table.insert(names, \"trigger\")\n"
"    end\n"
"    if bit.band(privilege, box.priv.INSERT) ~= 0 then\n"
"        table.insert(names, \"insert\")\n"
"    end\n"
"    if bit.band(privilege, box.priv.UPDATE) ~= 0 then\n"
"        table.insert(names, \"update\")\n"
"    end\n"
"    if bit.band(privilege, box.priv.DELETE) ~= 0 then\n"
"        table.insert(names, \"delete\")\n"
"    end\n"
"    return table.concat(names, \",\")\n"
"end\n"
"\n"
"local function object_resolve(object_type, object_name)\n"
"    if object_name ~= nil and type(object_name) ~= 'string'\n"
"            and type(object_name) ~= 'number' then\n"
"        box.error(box.error.ILLEGAL_PARAMS, \"wrong object name type\")\n"
"    end\n"
"    if object_type == 'universe' then\n"
"        return 0\n"
"    end\n"
"    if object_type == 'space' then\n"
"        if object_name == '' then\n"
"            return ''\n"
"        end\n"
"        local space = box.space[object_name]\n"
"        if  space == nil then\n"
"            box.error(box.error.NO_SUCH_SPACE, object_name)\n"
"        end\n"
"        return space.id\n"
"    end\n"
"    if object_type == 'function' then\n"
"        if object_name == '' then\n"
"            return ''\n"
"        end\n"
"        local _vfunc = box.space[box.schema.VFUNC_ID]\n"
"        local func\n"
"        if type(object_name) == 'string' then\n"
"            func = _vfunc.index.name:get{object_name}\n"
"        else\n"
"            func = _vfunc:get{object_name}\n"
"        end\n"
"        if func then\n"
"            return func[1]\n"
"        else\n"
"            box.error(box.error.NO_SUCH_FUNCTION, object_name)\n"
"        end\n"
"    end\n"
"    if object_type == 'sequence' then\n"
"        if object_name == '' then\n"
"            return ''\n"
"        end\n"
"        local seq = sequence_resolve(object_name)\n"
"        if seq == nil then\n"
"            box.error(box.error.NO_SUCH_SEQUENCE, object_name)\n"
"        end\n"
"        return seq\n"
"    end\n"
"    if object_type == 'role' or object_type == 'user' then\n"
"        if object_name == '' then\n"
"            return ''\n"
"        end\n"
"        local _vuser = box.space[box.schema.VUSER_ID]\n"
"        local role_or_user\n"
"        if type(object_name) == 'string' then\n"
"            role_or_user = _vuser.index.name:get{object_name}\n"
"        else\n"
"            role_or_user = _vuser:get{object_name}\n"
"        end\n"
"        if role_or_user and role_or_user[4] == object_type then\n"
"            return role_or_user[1]\n"
"        elseif object_type == 'role' then\n"
"            box.error(box.error.NO_SUCH_ROLE, object_name)\n"
"        else\n"
"            box.error(box.error.NO_SUCH_USER, object_name)\n"
"        end\n"
"    end\n"
"\n"
"    box.error(box.error.UNKNOWN_SCHEMA_OBJECT, object_type)\n"
"end\n"
"\n"
"local function object_name(object_type, object_id)\n"
"    if object_type == 'universe' or object_id == '' then\n"
"        return \"\"\n"
"    end\n"
"    local space\n"
"    if object_type == 'space' then\n"
"        space = box.space._vspace\n"
"    elseif object_type == 'sequence' then\n"
"        space = box.space._sequence\n"
"    elseif object_type == 'function' then\n"
"        space = box.space._vfunc\n"
"    elseif object_type == 'role' or object_type == 'user' then\n"
"        space = box.space._vuser\n"
"    else\n"
"        box.error(box.error.UNKNOWN_SCHEMA_OBJECT, object_type)\n"
"    end\n"
"    return space:get{object_id}[3]\n"
"end\n"
"\n"
"box.schema.func = {}\n"
"box.schema.func.create = function(name, opts)\n"
"    opts = opts or {}\n"
"    check_param_table(opts, { setuid = 'boolean',\n"
"                              if_not_exists = 'boolean',\n"
"                              language = 'string'})\n"
"    local _func = box.space[box.schema.FUNC_ID]\n"
"    local _vfunc = box.space[box.schema.VFUNC_ID]\n"
"    local func = _vfunc.index.name:get{name}\n"
"    if func then\n"
"        if not opts.if_not_exists then\n"
"            box.error(box.error.FUNCTION_EXISTS, name)\n"
"        end\n"
"        return\n"
"    end\n"
"    opts = update_param_table(opts, { setuid = false, language = 'lua'})\n"
"    opts.language = string.upper(opts.language)\n"
"    opts.setuid = opts.setuid and 1 or 0\n"
"    _func:auto_increment{session.euid(), name, opts.setuid, opts.language}\n"
"end\n"
"\n"
"box.schema.func.drop = function(name, opts)\n"
"    opts = opts or {}\n"
"    check_param_table(opts, { if_exists = 'boolean' })\n"
"    local _func = box.space[box.schema.FUNC_ID]\n"
"    local _vfunc = box.space[box.schema.VFUNC_ID]\n"
"    local fid\n"
"    local tuple\n"
"    if type(name) == 'string' then\n"
"        tuple = _vfunc.index.name:get{name}\n"
"    else\n"
"        tuple = _vfunc:get{name}\n"
"    end\n"
"    if tuple then\n"
"        fid = tuple[1]\n"
"    end\n"
"    if fid == nil then\n"
"        if not opts.if_exists then\n"
"            box.error(box.error.NO_SUCH_FUNCTION, name)\n"
"        end\n"
"        return\n"
"    end\n"
"    revoke_object_privs('function', fid)\n"
"    _func:delete{fid}\n"
"end\n"
"\n"
"function box.schema.func.exists(name_or_id)\n"
"    local _vfunc = box.space[box.schema.VFUNC_ID]\n"
"    local tuple = nil\n"
"    if type(name_or_id) == 'string' then\n"
"        tuple = _vfunc.index.name:get{name_or_id}\n"
"    elseif type(name_or_id) == 'number' then\n"
"        tuple = _vfunc:get{name_or_id}\n"
"    end\n"
"    return tuple ~= nil\n"
"end\n"
"\n"
"box.schema.func.reload = internal.func_reload\n"
"\n"
"box.internal.collation = {}\n"
"box.internal.collation.create = function(name, coll_type, locale, opts)\n"
"    opts = opts or setmap{}\n"
"    if type(name) ~= 'string' then\n"
"        box.error(box.error.ILLEGAL_PARAMS,\n"
"        \"name (first arg) must be a string\")\n"
"    end\n"
"    if type(coll_type) ~= 'string' then\n"
"        box.error(box.error.ILLEGAL_PARAMS,\n"
"        \"type (second arg) must be a string\")\n"
"    end\n"
"    if type(locale) ~= 'string' then\n"
"        box.error(box.error.ILLEGAL_PARAMS,\n"
"        \"locale (third arg) must be a string\")\n"
"    end\n"
"    if type(opts) ~= 'table' then\n"
"        box.error(box.error.ILLEGAL_PARAMS,\n"
"        \"options (fourth arg) must be a table or nil\")\n"
"    end\n"
"    local lua_opts = {if_not_exists = opts.if_not_exists }\n"
"    check_param_table(lua_opts, {if_not_exists = 'boolean'})\n"
"    opts.if_not_exists = nil\n"
"    opts = setmap(opts)\n"
"\n"
"    local _coll = box.space[box.schema.COLLATION_ID]\n"
"    if lua_opts.if_not_exists then\n"
"        local coll = _coll.index.name:get{name}\n"
"        if coll then\n"
"            return\n"
"        end\n"
"    end\n"
"    _coll:auto_increment{name, session.euid(), coll_type, locale, opts}\n"
"end\n"
"\n"
"box.internal.collation.drop = function(name, opts)\n"
"    opts = opts or {}\n"
"    check_param_table(opts, { if_exists = 'boolean' })\n"
"\n"
"    local _coll = box.space[box.schema.COLLATION_ID]\n"
"    if opts.if_exists then\n"
"        local coll = _coll.index.name:get{name}\n"
"        if not coll then\n"
"            return\n"
"        end\n"
"    end\n"
"    _coll.index.name:delete{name}\n"
"end\n"
"\n"
"box.internal.collation.exists = function(name)\n"
"    local _coll = box.space[box.schema.COLLATION_ID]\n"
"    local coll = _coll.index.name:get{name}\n"
"    return not not coll\n"
"end\n"
"\n"
"box.internal.collation.id_by_name = function(name)\n"
"    local _coll = box.space[box.schema.COLLATION_ID]\n"
"    local coll = _coll.index.name:get{name}\n"
"    return coll[1]\n"
"end\n"
"\n"
"box.schema.user = {}\n"
"\n"
"box.schema.user.password = function(password)\n"
"    local BUF_SIZE = 128\n"
"    local buf = ffi.new(\"char[\?]\", BUF_SIZE)\n"
"    builtin.password_prepare(password, #password, buf, BUF_SIZE)\n"
"    return ffi.string(buf)\n"
"end\n"
"\n"
"local function chpasswd(uid, new_password)\n"
"    local _user = box.space[box.schema.USER_ID]\n"
"    local auth_mech_list = {}\n"
"    auth_mech_list[\"chap-sha1\"] = box.schema.user.password(new_password)\n"
"    _user:update({uid}, {{\"=\", 5, auth_mech_list}})\n"
"end\n"
"\n"
"box.schema.user.passwd = function(name, new_password)\n"
"    if name == nil then\n"
"        box.error(box.error.PROC_LUA, \"Usage: box.schema.user.passwd([user,] password)\")\n"
"    end\n"
"    if new_password == nil then\n"
"        -- change password for current user\n"
"        new_password = name\n"
"        box.session.su('admin', chpasswd, session.uid(), new_password)\n"
"    else\n"
"        -- change password for other user\n"
"        local uid = user_resolve(name)\n"
"        if uid == nil then\n"
"            box.error(box.error.NO_SUCH_USER, name)\n"
"        end\n"
"        return chpasswd(uid, new_password)\n"
"    end\n"
"end\n"
"\n"
"box.schema.user.create = function(name, opts)\n"
"    local uid = user_or_role_resolve(name)\n"
"    opts = opts or {}\n"
"    check_param_table(opts, { password = 'string', if_not_exists = 'boolean' })\n"
"    if uid then\n"
"        if not opts.if_not_exists then\n"
"            box.error(box.error.USER_EXISTS, name)\n"
"        end\n"
"        return\n"
"    end\n"
"    local auth_mech_list = setmap({})\n"
"    if opts.password then\n"
"        auth_mech_list[\"chap-sha1\"] = box.schema.user.password(opts.password)\n"
"    end\n"
"    local _user = box.space[box.schema.USER_ID]\n"
"    uid = _user:auto_increment{session.euid(), name, 'user', auth_mech_list}[1]\n"
"    -- grant role 'public' to the user\n"
"    box.schema.user.grant(uid, 'public')\n"
"    -- Grant privilege 'alter' on itself, so that it can\n"
"    -- change its password or username.\n"
"    box.schema.user.grant(uid, 'alter', 'user', uid)\n"
"    -- we have to grant global privileges from setuid function, since\n"
"    -- only admin has the ownership over universe and we don't have\n"
"    -- grant option\n"
"    box.session.su('admin', box.schema.user.grant, uid, 'session,usage', 'universe',\n"
"                   nil, {if_not_exists=true})\n"
"end\n"
"\n"
"box.schema.user.exists = function(name)\n"
"    if user_resolve(name) then\n"
"        return true\n"
"    else\n"
"        return false\n"
"    end\n"
"end\n"
"\n"
"local function grant(uid, name, privilege, object_type,\n"
"                     object_name, options)\n"
"    -- From user point of view, role is the same thing\n"
"    -- as a privilege. Allow syntax grant(user, role).\n"
"    if object_name == nil then\n"
"        if object_type == nil then\n"
"            -- sic: avoid recursion, to not bother with roles\n"
"            -- named 'execute'\n"
"            object_type = 'role'\n"
"            object_name = privilege\n"
"            privilege = 'execute'\n"
"        else\n"
"            -- Allow syntax grant(user, priv, entity)\n"
"            -- for entity grants.\n"
"            object_name = ''\n"
"        end\n"
"    end\n"
"    local privilege_hex = privilege_check(privilege, object_type)\n"
"\n"
"    local oid = object_resolve(object_type, object_name)\n"
"    options = options or {}\n"
"    if options.grantor == nil then\n"
"        options.grantor = session.euid()\n"
"    else\n"
"        options.grantor = user_or_role_resolve(options.grantor)\n"
"    end\n"
"    local _priv = box.space[box.schema.PRIV_ID]\n"
"    local _vpriv = box.space[box.schema.VPRIV_ID]\n"
"    -- add the granted privilege to the current set\n"
"    local tuple = _vpriv:get{uid, object_type, oid}\n"
"    local old_privilege\n"
"    if tuple ~= nil then\n"
"        old_privilege = tuple[5]\n"
"    else\n"
"        old_privilege = 0\n"
"    end\n"
"    privilege_hex = bit.bor(privilege_hex, old_privilege)\n"
"    -- do not execute a replace if it does not change anything\n"
"    -- XXX bug if we decide to add a grant option: new grantor\n"
"    -- replaces the old one, old grantor is lost\n"
"    if privilege_hex ~= old_privilege then\n"
"        _priv:replace{options.grantor, uid, object_type, oid, privilege_hex}\n"
"    elseif not options.if_not_exists then\n"
"            if object_type == 'role' and object_name ~= '' and\n"
"               privilege == 'execute' then\n"
"                box.error(box.error.ROLE_GRANTED, name, object_name)\n"
"            else\n"
"                box.error(box.error.PRIV_GRANTED, name, privilege,\n"
"                          object_type, object_name)\n"
"            end\n"
"    end\n"
"end\n"
"\n"
"local function revoke(uid, name, privilege, object_type, object_name, options)\n"
"    -- From user point of view, role is the same thing\n"
"    -- as a privilege. Allow syntax revoke(user, role).\n"
"    if object_name == nil then\n"
"        if object_type == nil then\n"
"            object_type = 'role'\n"
"            object_name = privilege\n"
"            privilege = 'execute'\n"
"        else\n"
"            -- Allow syntax revoke(user, privilege, entity)\n"
"            -- to revoke entity privileges.\n"
"            object_name = ''\n"
"        end\n"
"    end\n"
"    local privilege_hex = privilege_check(privilege, object_type)\n"
"    options = options or {}\n"
"    local oid = object_resolve(object_type, object_name)\n"
"    local _priv = box.space[box.schema.PRIV_ID]\n"
"    local _vpriv = box.space[box.schema.VPRIV_ID]\n"
"    local tuple = _vpriv:get{uid, object_type, oid}\n"
"    -- system privileges of admin and guest can't be revoked\n"
"    if tuple == nil then\n"
"        if options.if_exists then\n"
"            return\n"
"        end\n"
"        if object_type == 'role' and object_name ~= '' and\n"
"           privilege == 'execute' then\n"
"            box.error(box.error.ROLE_NOT_GRANTED, name, object_name)\n"
"        else\n"
"            box.error(box.error.PRIV_NOT_GRANTED, name, privilege,\n"
"                      object_type, object_name)\n"
"        end\n"
"    end\n"
"    local old_privilege = tuple[5]\n"
"    local grantor = tuple[1]\n"
"    -- sic:\n"
"    -- a user may revoke more than he/she granted\n"
"    -- (erroneous user input)\n"
"    --\n"
"    privilege_hex = bit.band(old_privilege, bit.bnot(privilege_hex))\n"
"    -- give an error if we're not revoking anything\n"
"    if privilege_hex == old_privilege then\n"
"        if options.if_exists then\n"
"            return\n"
"        end\n"
"        box.error(box.error.PRIV_NOT_GRANTED, name, privilege,\n"
"                  object_type, object_name)\n"
"    end\n"
"    if privilege_hex ~= 0 then\n"
"        _priv:replace{grantor, uid, object_type, oid, privilege_hex}\n"
"    else\n"
"        _priv:delete{uid, object_type, oid}\n"
"    end\n"
"end\n"
"\n"
"local function drop(uid, opts)\n"
"    -- recursive delete of user data\n"
"    local _vpriv = box.space[box.schema.VPRIV_ID]\n"
"    local spaces = box.space[box.schema.VSPACE_ID].index.owner:select{uid}\n"
"    for k, tuple in pairs(spaces) do\n"
"        box.space[tuple[1]]:drop()\n"
"    end\n"
"    local funcs = box.space[box.schema.VFUNC_ID].index.owner:select{uid}\n"
"    for k, tuple in pairs(funcs) do\n"
"        box.schema.func.drop(tuple[1])\n"
"    end\n"
"    -- if this is a role, revoke this role from whoever it was granted to\n"
"    local grants = _vpriv.index.object:select{'role', uid}\n"
"    for k, tuple in pairs(grants) do\n"
"        revoke(tuple[2], tuple[2], uid)\n"
"    end\n"
"    local sequences = box.space[box.schema.VSEQUENCE_ID].index.owner:select{uid}\n"
"    for k, tuple in pairs(sequences) do\n"
"        box.schema.sequence.drop(tuple[1])\n"
"    end\n"
"    -- xxx: hack, we have to revoke session and usage privileges\n"
"    -- of a user using a setuid function in absence of create/drop\n"
"    -- privileges and grant option\n"
"    if box.space._vuser:get{uid}[4] == 'user' then\n"
"        box.session.su('admin', box.schema.user.revoke, uid,\n"
"                       'session,usage', 'universe', nil, {if_exists = true})\n"
"    end\n"
"    local privs = _vpriv.index.primary:select{uid}\n"
"\n"
"    for k, tuple in pairs(privs) do\n"
"        -- we need an additional box.session.su() here, because of\n"
"        -- unnecessary check for privilege PRIV_REVOKE in priv_def_check()\n"
"        box.session.su(\"admin\", revoke, uid, uid, tuple[5], tuple[3], tuple[4])\n"
"    end\n"
"    box.space[box.schema.USER_ID]:delete{uid}\n"
"end\n"
"\n"
"box.schema.user.grant = function(user_name, ...)\n"
"    local uid = user_resolve(user_name)\n"
"    if uid == nil then\n"
"        box.error(box.error.NO_SUCH_USER, user_name)\n"
"    end\n"
"    return grant(uid, user_name, ...)\n"
"end\n"
"\n"
"box.schema.user.revoke = function(user_name, ...)\n"
"    local uid = user_resolve(user_name)\n"
"    if uid == nil then\n"
"        box.error(box.error.NO_SUCH_USER, user_name)\n"
"    end\n"
"    return revoke(uid, user_name, ...)\n"
"end\n"
"\n"
"box.schema.user.enable = function(user)\n"
"    box.schema.user.grant(user, \"session,usage\", \"universe\", nil,\n"
"                            {if_not_exists = true})\n"
"end\n"
"\n"
"box.schema.user.disable = function(user)\n"
"    box.schema.user.revoke(user, \"session,usage\", \"universe\", nil,\n"
"                            {if_exists = true})\n"
"end\n"
"\n"
"box.schema.user.drop = function(name, opts)\n"
"    opts = opts or {}\n"
"    check_param_table(opts, { if_exists = 'boolean' })\n"
"    local uid = user_resolve(name)\n"
"    if uid ~= nil then\n"
"        if uid >= box.schema.SYSTEM_USER_ID_MIN and\n"
"           uid <= box.schema.SYSTEM_USER_ID_MAX then\n"
"            -- gh-1205: box.schema.user.info fails\n"
"            box.error(box.error.DROP_USER, name,\n"
"                      \"the user or the role is a system\")\n"
"        end\n"
"        if uid == box.session.uid() or uid == box.session.euid() then\n"
"            box.error(box.error.DROP_USER, name,\n"
"                      \"the user is active in the current session\")\n"
"        end\n"
"        return drop(uid, opts)\n"
"    end\n"
"    if not opts.if_exists then\n"
"        box.error(box.error.NO_SUCH_USER, name)\n"
"    end\n"
"    return\n"
"end\n"
"\n"
"local function info(id)\n"
"    local _priv = box.space._vpriv\n"
"    local privs = {}\n"
"    for _, v in pairs(_priv:select{id}) do\n"
"        table.insert(\n"
"            privs,\n"
"            {privilege_name(v[5]), v[3], object_name(v[3], v[4])}\n"
"        )\n"
"    end\n"
"    return privs\n"
"end\n"
"\n"
"box.schema.user.info = function(user_name)\n"
"    local uid\n"
"    if user_name == nil then\n"
"        uid = box.session.euid()\n"
"    else\n"
"        uid = user_resolve(user_name)\n"
"        if uid == nil then\n"
"            box.error(box.error.NO_SUCH_USER, user_name)\n"
"        end\n"
"    end\n"
"    return info(uid)\n"
"end\n"
"\n"
"box.schema.role = {}\n"
"\n"
"box.schema.role.exists = function(name)\n"
"    if role_resolve(name) then\n"
"        return true\n"
"    else\n"
"        return false\n"
"    end\n"
"end\n"
"\n"
"box.schema.role.create = function(name, opts)\n"
"    opts = opts or {}\n"
"    check_param_table(opts, { if_not_exists = 'boolean' })\n"
"    local uid = user_or_role_resolve(name)\n"
"    if uid then\n"
"        if not opts.if_not_exists then\n"
"            box.error(box.error.ROLE_EXISTS, name)\n"
"        end\n"
"        return\n"
"    end\n"
"    local _user = box.space[box.schema.USER_ID]\n"
"    _user:auto_increment{session.euid(), name, 'role', setmap({})}\n"
"end\n"
"\n"
"box.schema.role.drop = function(name, opts)\n"
"    opts = opts or {}\n"
"    check_param_table(opts, { if_exists = 'boolean' })\n"
"    local uid = role_resolve(name)\n"
"    if uid == nil then\n"
"        if not opts.if_exists then\n"
"            box.error(box.error.NO_SUCH_ROLE, name)\n"
"        end\n"
"        return\n"
"    end\n"
"    if uid >= box.schema.SYSTEM_USER_ID_MIN and\n"
"       uid <= box.schema.SYSTEM_USER_ID_MAX or uid == box.schema.SUPER_ROLE_ID then\n"
"        -- gh-1205: box.schema.user.info fails\n"
"        box.error(box.error.DROP_USER, name, \"the user or the role is a system\")\n"
"    end\n"
"    return drop(uid)\n"
"end\n"
"\n"
"local function role_check_grant_revoke_of_sys_priv(priv)\n"
"    priv = string.lower(priv)\n"
"    if (type(priv) == 'string' and (priv:match(\"session\") or priv:match(\"usage\"))) or\n"
"        (type(priv) == \"number\" and (bit.band(priv, 8) ~= 0 or bit.band(priv, 16) ~= 0)) then\n"
"        box.error(box.error.GRANT, \"system privilege can not be granted to role\")\n"
"    end\n"
"end\n"
"\n"
"box.schema.role.grant = function(user_name, ...)\n"
"    local uid = role_resolve(user_name)\n"
"    if uid == nil then\n"
"        box.error(box.error.NO_SUCH_ROLE, user_name)\n"
"    end\n"
"    role_check_grant_revoke_of_sys_priv(...)\n"
"    return grant(uid, user_name, ...)\n"
"end\n"
"box.schema.role.revoke = function(user_name, ...)\n"
"    local uid = role_resolve(user_name)\n"
"    if uid == nil then\n"
"        box.error(box.error.NO_SUCH_ROLE, user_name)\n"
"    end\n"
"    role_check_grant_revoke_of_sys_priv(...)\n"
"    return revoke(uid, user_name, ...)\n"
"end\n"
"box.schema.role.info = function(role_name)\n"
"    local rid = role_resolve(role_name)\n"
"    if rid == nil then\n"
"        box.error(box.error.NO_SUCH_ROLE, role_name)\n"
"    end\n"
"    return info(rid)\n"
"end\n"
"\n"
"--\n"
"-- once\n"
"--\n"
"box.once = function(key, func, ...)\n"
"    if type(key) ~= 'string' or type(func) ~= 'function' then\n"
"        box.error(box.error.ILLEGAL_PARAMS, \"Usage: box.once(key, func, ...)\")\n"
"    end\n"
"\n"
"    local key = \"once\"..key\n"
"    if box.space._schema:get{key} ~= nil then\n"
"        return\n"
"    end\n"
"    box.ctl.wait_rw()\n"
"    box.space._schema:put{key}\n"
"    return func(...)\n"
"end\n"
"\n"
"--\n"
"-- nice output when typing box.space in admin console\n"
"--\n"
"box.space = {}\n"
"\n"
"local function box_space_mt(tab)\n"
"    local t = {}\n"
"    for k,v in pairs(tab) do\n"
"        -- skip system spaces and views\n"
"        if type(k) == 'string' and #k > 0 and k:sub(1,1) ~= '_' then\n"
"            t[k] = {\n"
"                engine = v.engine,\n"
"                is_local = v.is_local,\n"
"                temporary = v.temporary,\n"
"            }\n"
"        end\n"
"    end\n"
"    return t\n"
"end\n"
"\n"
"setmetatable(box.space, { __serialize = box_space_mt })\n"
"\n"
"box.internal.schema = {}\n"
"box.internal.schema.init = function()\n"
"    box_sequence_init()\n"
"end\n"
"\n"
"box.feedback = {}\n"
"box.feedback.save = function(file_name)\n"
"    if type(file_name) ~= \"string\" then\n"
"        error(\"Usage: box.feedback.save(path)\")\n"
"    end\n"
"    local feedback = json.encode(box.internal.feedback_daemon.generate_feedback())\n"
"    local fh, err = fio.open(file_name, {'O_CREAT', 'O_RDWR', 'O_TRUNC'},\n"
"        tonumber('0777', 8))\n"
"    if not fh then\n"
"        error(err)\n"
"    end\n"
"    fh:write(feedback)\n"
"    fh:close()\n"
"end\n"
"\n"
"box.NULL = msgpack.NULL\n"
""
;
